// created on 6/6/2004 at 10:46 AM
/*
 *   Copyright (c) 2004, Alexandros Frantzis (alf82 [at] freemail [dot] gr)
 *
 *   This file is part of Bless.
 *
 *   Bless is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   Bless is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Bless; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
 
using NUnit.Framework;
using System.Collections;
using System;
using System.IO;

namespace Bless.Buffers {

///<summary>
/// A buffer for holding bytes in a versatile manner.
/// It supports undo-redo and can easily handle large files.
/// Editing is also very cheap.
///</summary>
public class ByteBuffer {

	internal ChangeBuffer changeBuf;
	internal FileBuffer	fileBuf;
	internal SegmentCollection segCol;
	Stack undoStack;
	Stack redoStack;
	object SaveCheckpoint;
	
	// automatic file naming
	string autoFilename;
	static int autoNum=1;
	
	internal long size;
	
	// chaining related
	bool actionChaining;
	bool actionChainingFirst;
	
	public ByteBuffer() 
	{
		segCol=new SegmentCollection();
		changeBuf= new ChangeBuffer();
		undoStack= new Stack();
		redoStack= new Stack();
		size=0;
		SaveCheckpoint=null;
		
		// name the buffer automatically
		autoFilename="Untitled " + ByteBuffer.autoNum;
		ByteBuffer.autoNum++;	
	}
	
	///<summary>Create a ByteBuffer loaded with a file</summary>
	static public ByteBuffer FromFile(string filename)
	{
		ByteBuffer bb=new ByteBuffer();
		bb.LoadWithFileBuffer(new FileBuffer(filename, 0xfffff));
		
		// fix automatic file naming
		ByteBuffer.autoNum--;
		
		return bb;
	}
	
	///<summary>Create a ByteBuffer with a dummy-name</summary>
	public ByteBuffer(string filename): this() 
	{
		this.autoFilename=filename;
		
		// fix automatic file naming
		ByteBuffer.autoNum--;
	}
	
	///<summary>Regard all following actions as a single one</summary>
	public void BeginActionChaining()
	{
		actionChaining=true;
		actionChainingFirst=true;
	}
	
	///<summary>Stop regarding actions as a single one</summary>
	public void EndActionChaining()
	{
		actionChaining=false;
		actionChainingFirst=false;
		
		if (undoStack.Count > 0) {
			ByteBufferAction bba=(ByteBufferAction)undoStack.Peek();
			bba.ChainedNext=false;
		}
		
		if (Changed!=null)
			Changed(this);
	}
	
	///<summary>Set the chaining information of a ByteBufferAction</summary>
	void SetChaining(ByteBufferAction bba)
	{
		if (actionChainingFirst==true) {
			actionChainingFirst=false;
			bba.ChainedNext=true;
		}
		else if (actionChaining) {
			bba.ChainedPrevious=true;
			bba.ChainedNext=true;
		}
	}
	
	public delegate void ChangedHandler(ByteBuffer bb); 
	
	public event ChangedHandler Changed;
	
	///<summary>Append bytes at the end of the buffer</summary>
	public void Append(byte[] data) 
	{
		AppendAction aa=new AppendAction(data, this);	
		aa.Do();
		
		SetChaining(aa);
		undoStack.Push(aa);
		redoStack.Clear();
		if (!actionChaining && Changed!=null)
			Changed(this);
	}

	///<summary>Insert bytes into the buffer</summary>
	public void Insert(long pos, byte[] data) 
	{
		if (pos == size) {
			Append(data);
			return;
		}
		InsertAction ia=new InsertAction(pos, data, this);
		ia.Do();
		
		SetChaining(ia);		
		undoStack.Push(ia);
		redoStack.Clear();
		if (!actionChaining && Changed!=null)
			Changed(this);
	}		

	///<summary>Delete bytes from the buffer</summary>
	public void Delete(long pos1, long pos2) 
	{
		DeleteAction da=new DeleteAction(pos1, pos2, this);
		da.Do();
		
		SetChaining(da);		
		undoStack.Push(da);
		redoStack.Clear();
		if (!actionChaining && Changed!=null)
			Changed(this);
	}
	
	///<summary>Replace bytes in the buffer</summary>
	public void Replace(long pos1, long pos2, byte[] data) 
	{
		ReplaceAction ra=new ReplaceAction(pos1, pos2, data, this);
		ra.Do();
		
		SetChaining(ra);
		undoStack.Push(ra);
		redoStack.Clear();
		if (!actionChaining && Changed!=null)
			Changed(this);
	}
	
	///<summary>Undo the last action or chain of actions</summary>
	public void Undo() 
	{
		// while there are more actions
		while (undoStack.Count>0) {
			ByteBufferAction action=(ByteBufferAction)undoStack.Pop();
			action.Undo();
			redoStack.Push(action);
			
			// if this is the last action in the undo chain,
			// end the undo
			if (!action.ChainedPrevious) {
				if (Changed!=null)
					Changed(this);
				break;
			}
		}
	}
	
	///<summary>Redo the last undone action or chain of actions</summary>
	public void Redo() 
	{
		// while there are more actions
		while (redoStack.Count>0) {
			ByteBufferAction action=(ByteBufferAction)redoStack.Pop();
			action.Do();
			undoStack.Push(action);
			
			// if this is the last action in the redo chain,
			// end the redo (note the use of ChainedNext and 
			// not ChainedPrevious, as in Undo())
			if (!action.ChainedNext) {
				if (Changed!=null)
					Changed(this);
				break;
			}
		}
	}
	
	public class SaveInterruptedException : Exception { }
	
	// a delegate that is called while saving.
	// 'fraction' is the amount of job done eg 0.45=45%
	// 'show' whether to show the dialog (it is hidden by default)
	// the delegate returns whether the save should be canceled (true)
	// or may continue (false)   
	public delegate bool SaveProgressDelegate(double fraction, bool show);
	
	//<summary>method to call when save timer expires</summary> 
	void SaveTimerExpired(object o)
	{
		// make the SaveProgressDialog visible
		SaveProgressDelegate deleg=(SaveProgressDelegate)o;
		deleg(0.0, true);
	}
	
	
	///<summary>Save the buffer as a file</summary>
	public void SaveAs(string filename, bool reload, SaveProgressDelegate deleg)
	{
		SaveAs(filename, reload, deleg, true);
	}
	
	public void SaveAs(string filename, bool reload, SaveProgressDelegate deleg, bool setCheckPoint) 
	{
		// create 1 second timer
		System.Threading.Timer timer= new System.Threading.Timer(new System.Threading.TimerCallback(SaveTimerExpired), deleg, 1000, 0);
		FileStream fs=null;
		
		try {
			fs=new FileStream(filename, FileMode.Create, FileAccess.Write);
			const int blockSize=0xffff;
			
			byte[] baTemp=new byte[blockSize];
		
			// for every node
			Util.List.Node node=segCol.List.First;
			bool cancel=false;
			
			while (node!=null && !cancel)
			{
				// Save the data in the node 
				// in blocks of blockSize each
				Segment s=(Segment)node.data;
				long len=s.Size;
				long nBlocks=len/blockSize;
				int last=(int)(len%blockSize); // bytes in last block
				long i;
			
				// for every full block
				for (i=0;i<nBlocks;i++) {
					s.buffer.Get(baTemp, s.start+i*blockSize, blockSize);
					fs.Write(baTemp, 0, blockSize);
					if (deleg!=null) { // update and decide if we should cancel
						cancel=deleg(((double)((i+1)*blockSize))/len, false);
						if (cancel) break;	
					}
				}
			
				// if last non-full block is not empty
				if (last!=0 && !cancel) {
					s.buffer.Get(baTemp, s.start+i*blockSize, last);
					fs.Write(baTemp, 0, last);
				}
				
				if (deleg!=null) // update and decide if we should cancel
					cancel=deleg(((double)(i*blockSize))/len, false);
						
				node=node.next;
			}	
			
			fs.Close();
		
			// if save was canceled...
			if (cancel)
				throw new SaveInterruptedException();
		
			//Console.WriteLine("Saved to {0} {1}", filename, i);
		
			if (reload==true) {
				CloseFile();
				LoadWithFileBuffer(new FileBuffer(filename));
			}
		
			if (setCheckPoint) {
				if (undoStack.Count > 0)
					SaveCheckpoint=undoStack.Peek();
				else
					SaveCheckpoint=null;
			}
		}
		finally {
			timer.Dispose();
			if (fs!=null)
				fs.Close();
		}				
	}
	
	///<summary>Save the buffer under the same filename</summary>
	public void Save(SaveProgressDelegate deleg) 
	{
		string orig=this.Filename;
		if ((orig==null)||(orig==""))
			throw new System.ArgumentException("Filename is empty.");
			
		// 1. Save the buffer as a temporary file
		// 2. Delete the original file
		// 3. Move the temporary file to its proper position
		
		string tmp=Path.GetTempFileName();
		
		try {
			SaveAs(tmp, false, deleg, false);
			CloseFile();
			System.IO.File.Delete(orig); // this could throw
			System.IO.File.Move(tmp, orig);
			LoadWithFileBuffer(new FileBuffer(orig));
		}
		catch (SaveInterruptedException ex) {
			// delete temporary file
			System.IO.File.Delete(tmp);
			throw;
		}
		catch (Exception ex) {
			// reload the old file, but keep the changes
			// the user has made 
			fileBuf.Load(orig);
			throw;
		}
		
		if (undoStack.Count > 0)
			SaveCheckpoint=undoStack.Peek();
		else
			SaveCheckpoint=null;
	}
	
	///<summary> 
	/// Revert ByteBuffer to the last saved state
	///</summary> 
	public void Revert()
	{
		if (this.HasFile) {
			// reload file
			string filename=fileBuf.Filename;
			if (!File.Exists(filename))
				throw new FileNotFoundException(filename);
			
			fileBuf.Close();
			LoadWithFileBuffer(new FileBuffer(filename));

			undoStack.Clear();
			redoStack.Clear();
			SaveCheckpoint=null;
				
			// emit bytebuffer changed event
			if (Changed!=null)
				Changed(this);		
		}
	}
	
	///<summary> 
	/// Sets the file buffer and resets the segment collection
	///</summary> 
	private void LoadWithFileBuffer(FileBuffer fb)
	{
		fileBuf=fb;
		Segment s=new Segment(fileBuf, 0, fileBuf.Size-1);
		segCol=new SegmentCollection();
		segCol.Append(s);
		size=fileBuf.Size;
	}
	
	public byte this[long index] {
		set { } 
		get {
			long map; 
			Util.List.Node node;
			Segment seg=segCol.FindSegment(index, out map, out node);
			//Console.WriteLine("Searching index {0} at {1}:{2}", index, map, seg);
			if (seg==null)
				throw new IndexOutOfRangeException(string.Format("ByteBuffer[{0}]",index));
			else {
				try {
					return seg.buffer[seg.start+index-map];	
				}
				catch(IndexOutOfRangeException e) {
					Console.WriteLine("Problem at index {0} at {1}:{2}", index, map, seg);
					throw;
				}
			}
		}
	}
	
	
	public void CloseFile()
	{
		if (fileBuf != null)
				fileBuf.Close();
	} 
	
	public long Size {
		get { return size;}
	}

	public bool HasFile {
		get { return fileBuf != null; }
			
	}
	
	public string Filename {
		get {
			if (fileBuf != null)
				return fileBuf.Filename;
			else
				return this.autoFilename; 
			}	
	}

	public bool HasChanged {
		get {
			if (undoStack.Count>0)
				return (SaveCheckpoint != undoStack.Peek());
			else
				return SaveCheckpoint != null;
		}
	
	}
	
	public bool CanUndo {
		get { return (undoStack.Count>0); }
	}
	
	public bool CanRedo {
		get { return (redoStack.Count>0); }
	}
	
	internal void Display(string s) 
	{
		Console.Write(s);
		segCol.List.Display();
	}

}


#region Tests
[TestFixture]
public class ByteBufferTests {

	[Test]
	public void AppendDoTest() {
		ByteBuffer bb=new ByteBuffer();
		byte[] ba={1,3,5,7,13,17};
		bb.Append(ba);
		
		Assert.AreEqual(6, bb.Size);
		Assert.AreEqual(1, bb[0]);
		Assert.AreEqual(3, bb[1]);
		Assert.AreEqual(5, bb[2]);
		Assert.AreEqual(7, bb[3]);
		Assert.AreEqual(13, bb[4]);
		Assert.AreEqual(17, bb[5]);
	}
	
	[Test]
	public void AppendUndoTest() {
		ByteBuffer bb=new ByteBuffer();
		byte[] ba={1,3,5,7};
		byte[] ba1={13,17};
		bb.Append(ba);
		bb.Append(ba1);
			
		Assert.AreEqual(6, bb.Size,"#1");
		Assert.AreEqual(3, bb[1],"#2");
		Assert.AreEqual(17, bb[5],"#3");
		
		bb.Undo();
	
		Assert.AreEqual(4, bb.Size,"#4");
		Assert.AreEqual(1, bb[0],"#5");
		Assert.AreEqual(3, bb[1],"#6");
		Assert.AreEqual(5, bb[2],"#7");
		Assert.AreEqual(7, bb[3],"#8");
	
	}
	
	[Test]
	public void InsertDoTest() {
		ByteBuffer bb=new ByteBuffer();
		byte[] ba={1,3,17};
		byte[] ba1={5,7,13};
		bb.Append(ba);
		//bb.Display("D1");
		bb.Insert(2, ba1);
		//bb.Display("D2");
		
		Assert.AreEqual(6, bb.Size,"#1");
		Assert.AreEqual(1, bb[0],"#2");
		Assert.AreEqual(3, bb[1],"#3");
		Assert.AreEqual(5, bb[2],"#4");
		Assert.AreEqual(7, bb[3],"#5");
		Assert.AreEqual(13, bb[4],"#6");
		Assert.AreEqual(17, bb[5],"#7");
	}
	
	[Test]
	public void InsertUndoTest() {
		ByteBuffer bb=new ByteBuffer();
		byte[] ba={1,3,17};
		byte[] ba1={5,7,13};
		bb.Append(ba);
		//bb.Display("D1");
		bb.Insert(2, ba1);
		//bb.Display("D2");
	
		Assert.AreEqual(6, bb.Size,"#1");
		Assert.AreEqual(3, bb[1],"#2");
		Assert.AreEqual(17, bb[5],"#3");
		
		bb.Undo();
		//bb.Display("D3");
		Assert.AreEqual(3, bb.Size,"#4");
		Assert.AreEqual(1, bb[0],"#5");
		Assert.AreEqual(3, bb[1],"#6");
		Assert.AreEqual(17, bb[2],"#7");
	}

	[Test]
	public void DeleteDoTest() {
		ByteBuffer bb=new ByteBuffer();
		byte[] ba={1,3,17};
		byte[] ba1={5,7,13};
		bb.Append(ba);
		//bb.Display("D1");
		bb.Insert(2, ba1);
		//bb.Display("D2");
		bb.Delete(1,3);
		
		Assert.AreEqual(3, bb.Size,"#1");
		Assert.AreEqual(1, bb[0],"#2");
		Assert.AreEqual(13, bb[1],"#3");
		Assert.AreEqual(17, bb[2],"#4");
	}
	
	[Test]
	public void DeleteUndoTest() {
		ByteBuffer bb=new ByteBuffer();
		byte[] ba={1,3,17};
		byte[] ba1={5,7,13};
		bb.Append(ba);
		//bb.Display("D1");
		bb.Insert(2, ba1);
		//bb.Display("D2");
		bb.Delete(1,3);
		
		Assert.AreEqual(3, bb.Size,"#1");
		Assert.AreEqual(1, bb[0],"#2");
		Assert.AreEqual(13, bb[1],"#3");
		Assert.AreEqual(17, bb[2],"#4");
		
		bb.Undo();
		
		Assert.AreEqual(6, bb.Size,"#5");
		Assert.AreEqual(1, bb[0],"#6");
		Assert.AreEqual(3, bb[1],"#7");
		Assert.AreEqual(5, bb[2],"#8");
		Assert.AreEqual(7, bb[3],"#9");
		Assert.AreEqual(13, bb[4],"#10");
		Assert.AreEqual(17, bb[5],"#11");
		
	}

	[Test]
	public void FileLoadTest1() {
		ByteBuffer bb=new ByteBuffer("test2.bin");
	
		long size=bb.Size;
		int sum=0;
		
		for(int i=0; i<size; i++) 
			sum ^=bb[i];
	
		Assert.AreEqual( 0x3a, sum);
		
	}
	
	[Test]
	public void FileLoadTest2() {
		ByteBuffer bb=new ByteBuffer("test2.bin");
		byte[] ba={0x41, 0x61};
		bb.Append(ba);
		
		long size=bb.Size;
		int sum=0;
		
		for(int i=0; i<size; i++) 
			sum ^=bb[i];
	
		Assert.AreEqual( 0x3a^0x20, sum);
		
	}
	
	[Test]
	public void FileLoadTest3() {
		ByteBuffer bb=new ByteBuffer("test2.bin");
		byte[] ba={0x41, 0x61};
		bb.Insert(10000, ba);
		
		long size=bb.Size;
		int sum=0;
		
		for(int i=0; i<size; i++) 
			sum ^=bb[i];
	
		Assert.AreEqual( 0x3a^0x20, sum);
		
	}
	
	[Test]
	public void FileSaveTest1() {
		ByteBuffer bb=new ByteBuffer("test2.bin");
		byte[] ba={0x41, 0x61};
		long size1=bb.Size;
		
		bb.Insert(10000, ba);
		Assert.AreEqual(size1+2, bb.Size,"#1");
		
		bb.SaveAs("test2.bak", false, null);
		
		ByteBuffer bb1=new ByteBuffer("test2.bak");
		
		
		Assert.AreEqual(bb.Size, bb1.Size,"#2");
		long size=bb1.Size;
		
		for(int i=0; i<size; i++) { 
			if (bb1[i]!=bb[i])
				Assert.Fail(string.Format("#3 Difference at {0} ({1}!={2})",i, bb1[i], bb[i]));
		}
		
	}

	[Test]
	public void HasChangedTest() 
	{
		ByteBuffer bb=new ByteBuffer();
		byte[] ba={0x41, 0x61};
		Assert.AreEqual(false, bb.HasChanged,"#0");
		bb.Append(ba);
		Assert.AreEqual(true, bb.HasChanged,"#1");
		bb.Undo();
		Assert.AreEqual(false, bb.HasChanged,"#2");
	}

}

#endregion

}