// created on 6/14/2004 at 10:35 PM
/*
 *   Copyright (c) 2004, Alexandros Frantzis (alf82 [at] freemail [dot] gr)
 *
 *   This file is part of Bless.
 *
 *   Bless is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   Bless is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Bless; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
using System.Collections;
using Gtk;
using Bless.Buffers;
using Bless.Gui.Areas;
using Bless.Util;

namespace Bless.Gui {

///<summary>An interface to a widget that displays data from a buffer</summary>
public class DataView {
	Layout layout;
	ByteBuffer byteBuffer;
	Gtk.DrawingArea drawingArea;
	Gtk.VScrollbar vscroll;
	bool widgetRealized;
	
	DataViewControl dvControl;
	DataViewDisplay dvDisplay;
	
	byte[] clipboard;
	
	bool overwrite;
	
	//undo-redo related
	Stack cursorUndoStack;
	Stack cursorRedoStack;
	
	internal Stack CursorUndoStack {
		get { return cursorUndoStack; }
	}  
	
	internal Stack CursorRedoStack {
		get { return cursorRedoStack; }
	}
	
	public DataViewControl Control {
		get { return dvControl; }
	}
	
	public DataViewDisplay Display {
		get { return dvDisplay; }
	}
	
	public ByteBuffer Buffer {
		get { return byteBuffer; }
		set { 
			byteBuffer=value;
			byteBuffer.Changed+=OnByteBufferChanged;
			foreach(Area a in dvDisplay.Layout.Areas) {
				a.Buffer=byteBuffer;
				a.CursorOffset=0;
				a.CursorDigit=0;
				a.Selection.Clear();
			}
			dvDisplay.Redraw();
			dvDisplay.VScroll.Adjustment.Value=0; 
		}
	} 
	
	public bool Overwrite {
		set {
			overwrite=value; 
			if (OverwriteChanged!=null)
				OverwriteChanged(this);
		}
		
		get { return overwrite; }
	}
	
	public Area FocusedArea {
		get {
			Area a;
			dvControl.FindFocusedArea(out a);
			return a;
		}
	}
	
	private void OnByteBufferChanged(ByteBuffer bb)
	{
		//System.Console.WriteLine("ByteBuffer changed!");
		dvDisplay.Redraw();
	}
	
	private void DeleteSelectionInternal(Area area)
	{
		area.Selection.Sort();
		byteBuffer.Delete(area.Selection.Start, area.Selection.End);
		cursorUndoStack.Push(new CursorState(area.CursorOffset, 0, area.Selection.Start, 0));
		cursorRedoStack.Clear();
			
		//long bbSize=byteBuffer.Size;
		long cOffset=area.Selection.Start;
		
		this.MoveCursor(cOffset, 0);
		this.SetSelection(-1, -1);
	}
	
	///<summary>Create a DataView</summary>
	public DataView()
	{
		dvControl=new DataViewControl(this);
		dvDisplay=new DataViewDisplay(this);
		
		dvDisplay.Control=dvControl;
		dvControl.Display=dvDisplay;
		
		// initialize clipboard buffer
		clipboard=new byte[0];
		
		// initialize undo/redo cursor offset stacks
		cursorUndoStack=new Stack();
		cursorRedoStack=new Stack();
	}
	
	public void Copy()
	{
		if (dvDisplay.Layout.Areas.Count<=0)
			return;
		
		Area area0=((Area)dvDisplay.Layout.Areas[0]);
		area0.Selection.Sort();
		long clipSize=area0.Selection.Size;
		
		// if no selection, do nothing (keep old clipboard data)
		if (clipSize==0)
			return;
			
		clipboard=new byte[clipSize];
		
		long i=0;
		long start=area0.Selection.Start;
		while (i < clipSize) {
			clipboard[i]=byteBuffer[start+i];
			i++;
		}
		
		dvDisplay.MakeOffsetVisible(area0.CursorOffset, DataViewDisplay.ShowType.Closest);
	}
	
	public void Cut()
	{
		if (dvDisplay.Layout.Areas.Count<=0)
			return;
		
		Area area0=((Area)dvDisplay.Layout.Areas[0]);
		area0.Selection.Sort();
		long clipSize=area0.Selection.Size;
		
		// if no selection, do nothing (keep old clipboard data)
		if (clipSize==0)
			return;
		
		clipboard=new byte[clipSize];
		
		long i=0;
		long start=area0.Selection.Start;
		while (i < clipSize) {
			clipboard[i]=byteBuffer[start+i];
			i++;
		}
		
		byteBuffer.Delete(area0.Selection.Start, area0.Selection.End);
		cursorUndoStack.Push(new CursorState(area0.CursorOffset, 0, area0.Selection.Start, 0));
		cursorRedoStack.Clear();
		
		//set cursor		
		this.MoveCursor(area0.Selection.Start, 0);
			  
		//clear selections
		this.SetSelection(-1, -1);
		
		// Make sure dataView.Redraw() is called before dvDisplay.MakeOffsetVisible()
		// so that the Scrollbar has the correct range
		// when calling dataView.Goto(). 
		// dataView.Redraw();
		dvDisplay.MakeOffsetVisible(area0.CursorOffset, DataViewDisplay.ShowType.Closest);
	}
	
	public void Paste()
	{
		if (dvDisplay.Layout.Areas.Count<=0)
			return;
		
		Area area0=((Area)dvDisplay.Layout.Areas[0]);
		
		if (clipboard.Length == 0)
			return;
		
		// if no range is selected insert the data and move
		// cursor to the end of inserted data
		if (area0.Selection.IsEmpty()){
			byteBuffer.Insert(area0.CursorOffset, clipboard);
			cursorUndoStack.Push(new CursorState(area0.CursorOffset, 0, area0.CursorOffset+clipboard.Length, 0));
			cursorRedoStack.Clear();
			
			this.MoveCursor(area0.CursorOffset + clipboard.Length, 0);
		}	
		else {
			// if a range is selected, replace the range with
			// the data and move the cursor to the end of
			// the new data
			byteBuffer.Replace(area0.Selection.Start, area0.Selection.End, clipboard);
			cursorUndoStack.Push(new CursorState(area0.Selection.Start, 0, area0.Selection.Start+clipboard.Length, 0));
			cursorRedoStack.Clear();
			
			this.MoveCursor(area0.Selection.Start + clipboard.Length, 0);
			this.SetSelection(-1, -1);
		}
		
		// Make sure dataView.Redraw() is called before dvDisplay.MakeOffsetVisible()
		// so that the Scrollbar has the correct range
		// when calling dvDisplay.MakeOffsetVisible().
		// dataView.Redraw();
		dvDisplay.MakeOffsetVisible(area0.CursorOffset, DataViewDisplay.ShowType.Closest);
	}
	
	public void Delete()
	{
		if (dvDisplay.Layout.Areas.Count<=0)
			return;
		
		Area area0=((Area)dvDisplay.Layout.Areas[0]);
		
		// if nothing is selected, delete the byte at the current offset
		if (area0.Selection.IsEmpty()==true) {
			if (area0.CursorOffset < byteBuffer.Size) {
				byteBuffer.Delete(area0.CursorOffset, area0.CursorOffset);
				cursorUndoStack.Push(new CursorState(area0.CursorOffset, area0.CursorDigit, area0.CursorOffset, area0.CursorDigit));
				cursorRedoStack.Clear();
			}
		}
		else { // delete the selection 
			DeleteSelectionInternal(area0);
		}
		
		dvDisplay.MakeOffsetVisible(area0.CursorOffset, DataViewDisplay.ShowType.Closest);		
	}
	
	public void DeleteBackspace()
	{
		if (dvDisplay.Layout.Areas.Count<=0)
			return;
		
		Area area0=((Area)dvDisplay.Layout.Areas[0]);
		
		// if nothing is selected, delete the byte at the previous offset
		if (area0.Selection.IsEmpty()==true) {
			long cOffset=area0.CursorOffset;
			
			if (cOffset > 0) {
				byteBuffer.Delete(cOffset-1, cOffset-1);
				cursorUndoStack.Push(new CursorState(cOffset, area0.CursorDigit, cOffset-1, area0.CursorDigit));
				cursorRedoStack.Clear();
				this.MoveCursor(cOffset-1, area0.CursorDigit);
			}
		}
		else { // delete the selection 
			DeleteSelectionInternal(area0);
		}
		
		dvDisplay.MakeOffsetVisible(area0.CursorOffset, DataViewDisplay.ShowType.Closest);		
	}
	
	public void Undo()
	{
		// undo buffer changes
		byteBuffer.Undo();
		
		// restore cursor position
		if (cursorUndoStack.Count>0) {
			CursorState ch=(CursorState)cursorUndoStack.Pop();
			cursorRedoStack.Push(ch);
			
			// clear the selection
			this.SetSelection(-1, -1);
			
			dvDisplay.MakeOffsetVisible(ch.UndoOffset, DataViewDisplay.ShowType.Closest);
			this.MoveCursor(ch.UndoOffset, ch.UndoDigit);
		} 
	}
	
	public void Redo()
	{
		// redo buffer changes
		byteBuffer.Redo();
		
		// restore cursor position
		if (cursorRedoStack.Count>0) {
			CursorState ch=(CursorState)cursorRedoStack.Pop();
			cursorUndoStack.Push(ch);
			
			// clear the selection
			this.SetSelection(-1, -1);
			
			dvDisplay.MakeOffsetVisible(ch.UndoOffset, DataViewDisplay.ShowType.Closest);
			this.MoveCursor(ch.RedoOffset, ch.RedoDigit);
		} 
	}
	
	public void Revert()
	{
		byteBuffer.Revert();
		
		cursorRedoStack.Clear();
		cursorUndoStack.Clear();
		
		MoveCursor(0,0);
	}
	
	public void SetSelection(long start, long end)
	{
		foreach (Area a in dvDisplay.Layout.Areas) {
			a.SetSelection(start, end);
		}
		
		if (SelectionChanged!=null)
			SelectionChanged(this); 
	}
	
	public Bless.Util.Range Selection
	{
		get { 
			if (dvDisplay.Layout.Areas.Count<=0)
				return new Bless.Util.Range();
		
			Area area0=((Area)dvDisplay.Layout.Areas[0]);
			
			Bless.Util.Range r=area0.Selection;
			return new Bless.Util.Range(r); 
		}
		set { 
			this.SetSelection(value.Start, value.End);
		}
	}
	
	
	public void MoveCursor(long offset, int digit)
	{
		foreach (Area a in dvDisplay.Layout.Areas) {
			a.MoveCursor(offset, digit);
		}
		
		if (CursorChanged!=null)
			CursorChanged(this);
	}
	
	
	public long CursorOffset
	{
		get {
			if (dvDisplay.Layout.Areas.Count<=0)
				return -1;
		
			Area area0=((Area)dvDisplay.Layout.Areas[0]);
			
			return area0.CursorOffset;	
		}
		
		set { 
			this.MoveCursor(value, 0);   
		} 
	}
	
	public delegate void SelectionChangedHandler(DataView dv);
	public delegate void CursorChangedHandler(DataView dv);
	public delegate void OverwriteChangedHandler(DataView dv);
	
	// events
	public event SelectionChangedHandler SelectionChanged;
	public event CursorChangedHandler CursorChanged;
	public event OverwriteChangedHandler OverwriteChanged;

}// end DataView


// class to store cursor position for undo/redo
class CursorState {
	public long UndoOffset;
	public int UndoDigit;
	public long RedoOffset;
	public int RedoDigit;
	public CursorState(long uo, int ud, long ro, int rd) 
		{ UndoOffset=uo; RedoOffset=ro; UndoDigit=ud; RedoDigit=rd;}
}


}//namespace