// project created on 6/3/2004 at 8:34 PM
/*
 *   Copyright (c) 2004, Alexandros Frantzis (alf82 [at] freemail [dot] gr)
 *
 *   This file is part of Bless.
 *
 *   Bless is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   Bless is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Bless; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

using System;
using System.Collections;
using System.Threading;
using System.IO;
using Gtk;
using Glade;
using Bless.Buffers;
using Bless.Gui;
using Bless.Gui.Dialogs;
using Bless.Tools.Find;
using Bless.Util;

public class BlessMain
{
	[Glade.Widget] Gtk.HBox DataViewBox;
	[Glade.Widget] Gtk.Toolbar MainToolbar;
	[Glade.Widget] Gtk.Window MainWindow;
	
	[Glade.Widget] Gtk.Paned StatusPaned;
	[Glade.Widget] Gtk.Statusbar OffsetStatusbar;
	[Glade.Widget] Gtk.Statusbar SelectionStatusbar;
	[Glade.Widget] Gtk.Statusbar MessageStatusbar;
	[Glade.Widget] Gtk.Statusbar OverwriteStatusbar;
	[Glade.Widget] Gtk.Statusbar DummyStatusbar;
	
	[Glade.Widget] Gtk.MenuItem FileRevertMenuItem;
	[Glade.Widget] Gtk.MenuItem EditUndoMenuItem;
	[Glade.Widget] Gtk.MenuItem EditRedoMenuItem;
	[Glade.Widget] Gtk.CheckMenuItem ViewStatusbarShowMenuItem;
	[Glade.Widget] Gtk.CheckMenuItem ViewStatusbarOffsetMenuItem;
	[Glade.Widget] Gtk.CheckMenuItem ViewStatusbarSelectionMenuItem;
	[Glade.Widget] Gtk.CheckMenuItem ViewStatusbarOverwriteMenuItem;
	

	DataBook dataBook;
	FindDialog findDialog;
	ReplaceDialog replaceDialog;
	GotoOffsetDialog gotoOffsetDialog;
	
	IFinder finder;
	
	Timer StatusMessageTimer;
	
	Hashtable bufferHasChanged;
	
	public static void Main (string[] args)
	{
		new BlessMain (args);
	}

	public BlessMain (string[] args) 
	{
		Application.Init();
	
		// load main window from glade XML
		Glade.XML gxml = new Glade.XML (null, "bless.glade", "MainWindow", null);
		gxml.Autoconnect (this);
		
		// set the application icon
		MainWindow.Icon=new Gdk.Pixbuf("../data/bless-icon.png");
	
		// create timer for status bar messages
		StatusMessageTimer= new Timer(new TimerCallback(ClearStatusMessage), null, System.Threading.Timeout.Infinite, 0);
		
		// create a hash table to hold the status of each buffer
		// in the DataBook
		bufferHasChanged=new Hashtable();
		
		// create the DataBook
		dataBook=new DataBook();
		dataBook.SwitchPage+= new SwitchPageHandler(OnSwitchPage);
		
		DataViewBox.PackStart(dataBook);
		
		// create the find dialog (hidden)
		finder=new FileFinder(dataBook);
		finder.Strategy=new SimpleFindStrategy();
		findDialog=new FindDialog(finder);
		
		// create the replace dialog (hidden)
		replaceDialog=new ReplaceDialog(finder);
		
		// create the go to offset dialog (hidden)
		gotoOffsetDialog=new GotoOffsetDialog(dataBook);
		
		// create a buffer
		ByteBuffer bb=null;
	
		if (args.Length>0)
			bb=OpenFile(args[0]);
	
		if (bb==null) { 
			bb=new ByteBuffer();
			bufferHasChanged[bb]=true;
			bb.Changed+=OnBufferChanged;
		}
	
		// create and setup a  DataView
		DataView dv=CreateDataView(bb);
	
		// append the DataView to the DataBook
		string filename=GetBufferPath(bb);
		dataBook.AppendView(dv, new CloseViewDelegate(CloseFile), Path.GetFileName(filename));
	
		OnBufferChanged(bb);
	
		DataViewBox.ShowAll();
		Application.Run();
	}
	
	///<summary>Displays a message in the status bar for 5 sec</summary>
	void DisplayStatusMessage(string message)
	{
		MessageStatusbar.Pop(0);	
		MessageStatusbar.Push(0, message);
		StatusMessageTimer.Change(5000, 0);	
	}
	
	///<summary>Clears the status bar message (callback)</summary>
	void ClearStatusMessage(object o)
	{
		MessageStatusbar.Pop(0);	
		//MessageStatusbar.Push(0, message);
		StatusMessageTimer.Change(System.Threading.Timeout.Infinite, System.Threading.Timeout.Infinite);	
	}
	
	///<summary>
	/// Returns the ByteBuffer's file path or "Untitled" if path is empty
	///</summary>
	string GetBufferPath(ByteBuffer bb)
	{
		if (bb.Filename==null || bb.Filename.Length == 0)
			return "Untitled";
		else
			return bb.Filename;
	}
	
	///<summary>
	/// Create and setup a DataView 
	///</summary>
	DataView CreateDataView(ByteBuffer bb)
	{
		DataView dv=new DataView();
		dv.SelectionChanged += UpdateSelectionStatusbar; //OnDataViewSelectionChanged;
		dv.CursorChanged += UpdateCursorStatusbar; //OnDataViewCursorChanged;
		dv.OverwriteChanged += UpdateOverwriteStatusbar; //OnDataViewOverwriteChanged;
	
		dv.Buffer=bb;
		
		return dv;
	}
	
	 
	///<summary>
	/// Updates the window's title with data from the specified DataView,
	/// if the specified DataView is active
	///</summary>
	void UpdateWindowTitle(DataView dv)
	{
		DataView curdv=((DataViewDisplay)dataBook.CurrentPageWidget).View;
		ByteBuffer curbb=curdv.Buffer;
		
		ByteBuffer bb=dv.Buffer;
		string filename=GetBufferPath(bb);
		
		// if DataView is active
		if (curbb==bb) {
			if (bb.HasChanged)
				MainWindow.Title= filename + " * - Bless";
			else
				MainWindow.Title= filename + " - Bless";
		}
	}
	
	///<summary>
	/// Updates the title of the tab with the specified DataView
	///</summary>
	void UpdateTabLabel(DataView dv)
	{
		DataView curdv=((DataViewDisplay)dataBook.CurrentPageWidget).View;
		ByteBuffer curbb=curdv.Buffer;
	
		ByteBuffer bb=dv.Buffer;
		string filename=GetBufferPath(bb);
	
		if (bb.HasChanged)
			dataBook.SetTabLabelText(dv.Display, Path.GetFileName(filename) +"*");
		else
			dataBook.SetTabLabelText(dv.Display, Path.GetFileName(filename));
	
	}
	
	///<summary>
	/// Updates the undo/redo items with data from the specified DataView,
	/// if the specified DataView is active
	///</summary>
	void UpdateUndoRedo(DataView dv)
	{
		DataView curdv=((DataViewDisplay)dataBook.CurrentPageWidget).View;
		ByteBuffer curbb=curdv.Buffer;
	
		ByteBuffer bb=dv.Buffer;
	
		// if DataView is active
		if (curbb==bb) {
			// gray Undo if there is no undo action 
			if (bb.CanUndo)
				EditUndoMenuItem.Sensitive=true;
			else 
				EditUndoMenuItem.Sensitive=false;
	
			// gray Redo if there is no redo action	
			if (bb.CanRedo)
				EditRedoMenuItem.Sensitive=true;
			else
				EditRedoMenuItem.Sensitive=false;
		}	
	
	}
	
	///<summary>
	/// Updates the cursor status bar with data from the specified DataView,
	/// if the specified DataView is active
	///</summary>
	public void UpdateCursorStatusbar(DataView dv)
	{	
		DataView curdv=((DataViewDisplay)dataBook.CurrentPageWidget).View;
		if (curdv!=dv)
			return;
	
		long coffset=dv.CursorOffset;
		OffsetStatusbar.Pop(0);
		string str=string.Format("Offset: 0x{0:x} / 0x{1:x}", coffset, dv.Buffer.Size-1);
		OffsetStatusbar.Push(0, str);
	}
	
	///<summary>
	/// Updates the selection status bar with data from the specified DataView,
	/// if the specified DataView is active
	///</summary>
	public void UpdateSelectionStatusbar(DataView dv)
	{
		DataView curdv=((DataViewDisplay)dataBook.CurrentPageWidget).View;
		if (curdv!=dv)
			return;
	
		Bless.Util.Range sel=dv.Selection;
		sel.Sort();
		SelectionStatusbar.Pop(0);
		string str;
	
		if (sel.IsEmpty()==true)
			str="Selection: None";
		else	
			str=string.Format("Selection: 0x{0:x} to 0x{1:x} (0x{2:x} bytes)", sel.Start, sel.End, sel.Size);
	
		SelectionStatusbar.Push(0, str);
	}
	
	///<summary>
	/// Updates the overwrite status bar with data from the specified DataView,
	/// if the specified DataView is active
	///</summary>
	public void UpdateOverwriteStatusbar(DataView dv)
	{	
		DataView curdv=((DataViewDisplay)dataBook.CurrentPageWidget).View;
		if (curdv!=dv)
			return;
	
		OverwriteStatusbar.Pop(0);
		if (dv.Overwrite==true)
			OverwriteStatusbar.Push(0, " OVR");
		else
			OverwriteStatusbar.Push(0, " INS");
	}
	
	public void UpdateRevert(DataView dv)
	{
		DataView curdv=((DataViewDisplay)dataBook.CurrentPageWidget).View;
		if (curdv!=dv)
			return;
			
		if (dv.Buffer.HasFile)
			FileRevertMenuItem.Sensitive=true;
		else
			FileRevertMenuItem.Sensitive=false;
	}
	
	///<summary>Handle ByteBuffer changes</summary>
	void OnBufferChanged(ByteBuffer bb)
	{
		DataView dv;
		
		// find DataView that owns bb
		foreach (DataViewDisplay dvtemp in dataBook.Children) {
			if (dvtemp.View.Buffer==bb) {
				dv=dvtemp.View;
				break;	
			}
		}

		// if the status of bb has not changed since the
		// last invocation, there is no need to update
		// the titles
		if ((bool)bufferHasChanged[bb]!=bb.HasChanged) {
			bufferHasChanged[bb]=bb.HasChanged;
			UpdateWindowTitle(dv);
			UpdateTabLabel(dv);
		}
	
		// always update undo/redo and cursor status bar
		UpdateUndoRedo(dv);
		UpdateCursorStatusbar(dv);
	}
	
	///<summary>If file has changed ask whether to save it.
	/// Returns true if file saved or the user doesn't want to save it.</summary>
	private bool AskForSaveIfFileChanged(DataView dv)
	{
		ByteBuffer bb=dv.Buffer;
	
		if (bb == null)
			return true;
	
		if (bb.HasChanged) {
			SaveConfirmationAlert sca=new SaveConfirmationAlert(bb.Filename, MainWindow);
			ResponseType res=(ResponseType)sca.Run();
			sca.Destroy();
	
			if (res== ResponseType.Ok) {
				return SaveFile(dv, null, false);
			}
			else if (res==ResponseType.No){
				return true;
			}
			else /*if (res==ResponseType.Cancel)*/
				return false;
		}
	
		return true;
	}
	
	///<summary>If any file has changed ask whether to save it.
	/// Returns true if all changed files are saved or the user doesn't want to save them.</summary>
	private bool AskForSaveIfFilesChanged()
	{
		ArrayList list=new ArrayList();
	
		// create a list with the changed files
		int i=0;
		foreach (DataViewDisplay dv in dataBook.Children) {
			ByteBuffer bb=dv.View.Buffer;
			if (bb.HasChanged)
			list.Add(new SaveFileItem(true, bb.Filename, i));
			i++;
		}
	
		if (list.Count == 0)
			return true;
	
		// special handling if only one file changed
		if (list.Count == 1) {
			int page=((SaveFileItem)list[0]).Page;
			// make the page active 
			dataBook.Page=page;
		
			DataView dv=((DataViewDisplay)dataBook.GetNthPage(page)).View;
			return AskForSaveIfFileChanged(dv);
		}
	
		// show the confirmation alert 
		SaveFileItem[] array=(SaveFileItem[])list.ToArray(typeof(SaveFileItem));
	
		SaveConfirmationMultiAlert sca=new SaveConfirmationMultiAlert(array, MainWindow);
		ResponseType res=(ResponseType)sca.Run();
		sca.Destroy();
	
		// handle responses	
		if (res == ResponseType.Ok) {
			bool allSaved=true;
			
			// save the files the user specified
			foreach(SaveFileItem item in array) {
				if (item.Save==true) {
					DataView dv=((DataViewDisplay)dataBook.GetNthPage(item.Page)).View;
					// make page active
					dataBook.Page=item.Page;
					if (SaveFile(dv, null, false)==false)
						allSaved=false;
					else {
						UpdateTabLabel(dv);
						UpdateWindowTitle(dv);	
					}	
				}
			} 
	
			return allSaved;
		}
		else if (res==ResponseType.No){
			return true;
		}
		else /*if (res==ResponseType.Cancel)*/
			return false;
	}
	
	
	///<summary>
	/// Try to open the file as a ByteBuffer 
	///</summary>
	private ByteBuffer OpenFile(string filename)
	{
		string fullPath=null;
	
		// get the full path
		try {
			fullPath=Path.GetFullPath(filename);	
		}
		catch(Exception ex) {
			ErrorAlert ea=new ErrorAlert("Error opening file "+filename, ex.Message, MainWindow);
			ea.Run();
			ea.Destroy();
		}
	
		// if we can't get full path return
		if (fullPath==null)
			return null;
	
		try {	
			ByteBuffer bb=ByteBuffer.FromFile(fullPath);
			DisplayStatusMessage("Loaded file '"+ fullPath +"'");
			bb.Changed+=OnBufferChanged;
			bufferHasChanged[bb]=true;
			return bb;
		}
		catch(UnauthorizedAccessException ex) {
			ErrorAlert ea=new ErrorAlert("Error opening file '"+ fullPath+"'", "You do not have read permissions for the file you requested.", MainWindow);
			ea.Run();
			ea.Destroy();
		}
		catch(System.IO.FileNotFoundException ex) {
			ErrorAlert ea=new ErrorAlert("Error opening file '"+ fullPath+"'", "The file you requested does not exist.", MainWindow);
			ea.Run();
			ea.Destroy();
		}
		catch(System.IO.IOException ex) {
			ErrorAlert ea=new ErrorAlert("Error opening file '"+ fullPath+"'", ex.Message, MainWindow);
			ea.Run();
			ea.Destroy();
		}
		catch(System.ArgumentException ex) {
			ErrorAlert ea=new ErrorAlert("Error opening file '"+ fullPath+"'", ex.Message, MainWindow);
			ea.Run();
			ea.Destroy();
		}
		
		return null;
	}
	
	///<summary>
	/// Manage high-level file saving procedures (eg confirm file overwrites, get filenames etc) 
	///</summary>
	private bool SaveFile(DataView dv, string filename, bool forceSaveAs)
	{
		ByteBuffer bb=dv.Buffer;
	
	 
		// if a filename is given, save the file under the specified filename
		if (!forceSaveAs && filename!=null) 
			return SaveFileInternal(dv, filename);
	  
		// if a filename is not given but buffer
		// has a file associated with it, save it under the same filename
		if (!forceSaveAs && filename==null && bb.HasFile==true && bb.Filename.Length!=0)
			return SaveFileInternal(dv, bb.Filename);	  
	  
		// otherwise prompt user for a name	  
		Gtk.FileSelection fs=new Gtk.FileSelection("Save File As");
		bool done=false;
		bool fileSaved=true;
	
		do {
			ResponseType response = (ResponseType)fs.Run();
			fs.Hide();
			if (response == ResponseType.Ok) {
				// check to see whether file exists and prompt user to confirm
				if (File.Exists(fs.Filename)) {
					FileOverwriteAlert ea=new FileOverwriteAlert(fs.Filename, MainWindow);
					ResponseType response1 = (ResponseType)ea.Run();
					ea.Destroy();
					if (response1 == ResponseType.Ok) {
						fileSaved=SaveFileInternal(dv, fs.Filename);
						done=true;
					}
					else
						done=false;
				}
				else{ // !File.Exists(fs.Filename)
					fileSaved=SaveFileInternal(dv, fs.Filename);
					done=true;
				}
			}
			else { // response!=ResponseType.OK
				done=true;
				fileSaved=false;
			}
		} while(!done);
	
		fs.Destroy();	
		
		UpdateRevert(dv);
		
		return fileSaved; 
	}
	
	
	///<summary>
	/// Manage low-level file saving procedures 
	///</summary>
	private bool SaveFileInternal(DataView dv, string filename)
	{
		ByteBuffer bb=dv.Buffer;
	
		string fullPath=null;
	
		// don't allow dataview events while saving
		dv.Display.Sensitive=false;	
	
		// get the full path
		try {
			fullPath=Path.GetFullPath(filename);	
		}
		catch(Exception ex) {
			ErrorAlert ea=new ErrorAlert("Error saving file "+filename, ex.Message, MainWindow);
			ea.Run();
			ea.Destroy();
		}
	
		// if we can't get full path, return
		if (fullPath==null)
			return false;
	   
		SaveProgressDialog spd=null;
		try {
			// decide whether to Save or Save As
			if (fullPath!=bb.Filename) {
				string msg="Saving file '"+ bb.Filename +"' as '"+fullPath+"'";
				DisplayStatusMessage(msg+"...");
	
				spd=new SaveProgressDialog(msg);
				bb.SaveAs(fullPath, true, new ByteBuffer.SaveProgressDelegate(spd.Update));
				spd.Destroy();
	
				DisplayStatusMessage("The file has been saved as '" + fullPath +"'");
				MainWindow.Title=fullPath + "- Bless";
				dataBook.SetTabLabelText(dv.Display, Path.GetFileName(fullPath));
			}
			else {
				string msg="Saving file '"+fullPath+"'";
				DisplayStatusMessage(msg+"...");
				
				spd=new SaveProgressDialog(msg);
				bb.Save(new ByteBuffer.SaveProgressDelegate(spd.Update));
				spd.Destroy();
	
				DisplayStatusMessage("The file '" + fullPath + "' has been saved");
			}
			
			// update local bb entry status
			bufferHasChanged[bb]=bb.HasChanged;
	
			// re-allow DataView events 
			dv.Display.Sensitive=true;
	
			return true;
		}
		catch (UnauthorizedAccessException ex) {
			ErrorAlert ea=new ErrorAlert("Error saving file '"+ fullPath+"'", "You don't have write permission for the specified location.", MainWindow);
			ea.Run();
			ea.Destroy();		
		}
		catch (System.ArgumentException ex) {
			ErrorAlert ea=new ErrorAlert("Error saving file '"+ fullPath+"'", ex.Message, MainWindow);
			ea.Run();
			ea.Destroy();
		}
		catch (System.IO.IOException ex) {
			ErrorAlert ea=new ErrorAlert("Error saving file '"+ fullPath+"'", ex.Message, MainWindow);
			ea.Run();
			ea.Destroy();
		}
		catch(ByteBuffer.SaveInterruptedException ex) {
	
		}
	
		spd.Destroy();
	
		DisplayStatusMessage("The file '" + fullPath + "' has NOT been saved");
	
		// re-allow DataView events
		dv.Display.Sensitive=true;
	
		return false;
	}
	
	void CloseFile(DataView dv)
	{
		// if user decides not to close after all
		if (AskForSaveIfFileChanged(dv)==false)
			return;
	
		dv.Buffer.CloseFile();
		dataBook.RemovePage(dataBook.PageNum(dv.Display));
	
		// promptly free pixmap memory
		dv.Display.DisposePixmaps();
		dv=null;
		
		// Make sure the new current DataView has the focus
		DataViewDisplay dvd=(DataViewDisplay)dataBook.CurrentPageWidget;
		if (dvd!=null)
			dvd.GrabKeyboardFocus();
		else { 
			// there is no dataview left,
			// update title and statusbars
			MainWindow.Title="Bless - Gtk# Hex Editor";
			OffsetStatusbar.Pop(0);
			OffsetStatusbar.Push(0, "Offset: -");
			SelectionStatusbar.Pop(0);
			SelectionStatusbar.Push(0, "Selection: None");
			FileRevertMenuItem.Sensitive=false;
		}
	}
	
	///<summary>
	/// Callback for SwitchPage DataBook event 
	///</summary>
	void OnSwitchPage(object o, SwitchPageArgs args)
	{
		DataView dv=((DataViewDisplay)dataBook.GetNthPage((int)args.PageNum)).View;
	
		UpdateWindowTitle(dv);
		UpdateUndoRedo(dv);
		UpdateRevert(dv);
			
		UpdateCursorStatusbar(dv);
		UpdateSelectionStatusbar(dv);
		UpdateOverwriteStatusbar(dv);
	
		dv.Display.GrabKeyboardFocus();
	}
	
	///<summary>Assign the resize grip to a visible statusbar</summary>
	void AssignResizeGrip()
	{
		// clear all grips
		OffsetStatusbar.HasResizeGrip=false;
		SelectionStatusbar.HasResizeGrip=false;
		MessageStatusbar.HasResizeGrip=false;
		OverwriteStatusbar.HasResizeGrip=false;
		DummyStatusbar.Visible=false;
		
		// give resize grip to a statusbar
		if (OverwriteStatusbar.Visible==false){
			if (SelectionStatusbar.Visible==false) {
				if (OffsetStatusbar.Visible==false)
					DummyStatusbar.Visible=true; 
				else
					OffsetStatusbar.HasResizeGrip=true;
			}
			else
				SelectionStatusbar.HasResizeGrip=true;	
		}
		else
			OverwriteStatusbar.HasResizeGrip=true;
	}
	
	///<summary>Handle file->new command from menu</summary>
	public void OnFileNew(object o, EventArgs args) 
	{
		ByteBuffer bb=new ByteBuffer();
		bb.Changed+=OnBufferChanged;
		bufferHasChanged[bb]=true;
	
		// create and setup a  DataView
		DataView dv=CreateDataView(bb);
	
		dataBook.AppendView(dv, new CloseViewDelegate(CloseFile), Path.GetFileName(bb.Filename));
		OnBufferChanged(bb);
	}
	
	///<summary>Handle file->open command from menu</summary>
	public void OnFileOpen(object o, EventArgs args) 
	{
		// should we replace the current page or create a new one?
		bool replaceCurrentPage=dataBook.CanReplacePage(dataBook.CurrentPage);
		
		// Get path of file to open	
		Gtk.FileSelection fs=new Gtk.FileSelection("Open File");
		
		ResponseType result = (ResponseType)fs.Run();
		fs.Hide();
		
		// if user does not cancel
		if (result == ResponseType.Ok) {
			// try to open the file
			ByteBuffer bb=OpenFile(fs.Filename);

			// if open was successful
			if (bb!=null) {
				if (replaceCurrentPage) { // replace current page
					DataView dv=((DataViewDisplay)dataBook.CurrentPageWidget).View;
					dv.Buffer.CloseFile();
					dv.Buffer=bb;
					UpdateRevert(dv);
				}
				else { // create new page
					// create and setup a  DataView
					DataView dv=CreateDataView(bb);	
					dataBook.AppendView(dv, new CloseViewDelegate(CloseFile), Path.GetFileName(bb.Filename));	
				}
	
				OnBufferChanged(bb);
				
			}	
		}
	
		fs.Destroy();
	}
	
	///<summary>Handle file->save command from menu</summary>
	public void OnFileSave(object o, EventArgs args)
	{
		if (dataBook.NPages == 0)
			return;
	
		DataView dv=((DataViewDisplay)dataBook.CurrentPageWidget).View;
	
		SaveFile(dv, null, false);
		UpdateTabLabel(dv);
		UpdateWindowTitle(dv);	
	}
	
	///<summary>Handle file->save as command from menu</summary>
	public void OnFileSaveAs(object o, EventArgs args)
	{
		if (dataBook.NPages == 0)
			return;
	
		DataView dv=((DataViewDisplay)dataBook.CurrentPageWidget).View;
	
		SaveFile(dv, null, true);
		UpdateTabLabel(dv);
		UpdateWindowTitle(dv);	
	}
	
	///<summary>Handle file->revert command from menu</summary>
	public void OnFileRevert(object o, EventArgs args)
	{
		if (dataBook.NPages == 0)
			return;
	
		DataView dv=((DataViewDisplay)dataBook.CurrentPageWidget).View;
		
		RevertConfirmationAlert rca=new RevertConfirmationAlert(dv.Buffer.Filename, MainWindow);
		ResponseType response=(ResponseType)rca.Run();
		rca.Destroy();
		
		try {
			if (response==ResponseType.Ok)
				dv.Revert();
		}
		catch (FileNotFoundException ex) {
			ErrorAlert ea=new ErrorAlert("Error reverting file '"+ ex.Message+"'.","The file cannot be found. Perhaps it has been recently deleted." , MainWindow);
			ea.Run();
			ea.Destroy();
		}
	}
	
	///<summary>Handle file->close as command from menu</summary>
	public void OnFileClose(object o, EventArgs args)
	{
		if (dataBook.NPages == 0)
			return;
		DataView dv=((DataViewDisplay)dataBook.CurrentPageWidget).View;
	
		CloseFile(dv);
	}
	
	
	///<summary>Handle file->quit command from menu</summary>
	public void OnFileQuit(object o, EventArgs args) 
	{
		if (AskForSaveIfFilesChanged())
			Application.Quit ();
	}
	
	///<summary>Handle edit->undo command from menu</summary>
	public void OnEditUndo(object o, EventArgs args) 
	{
		if (dataBook.NPages == 0)
			return;
		DataView dv=((DataViewDisplay)dataBook.CurrentPageWidget).View;
		dv.Undo();
	}
	
	///<summary>Handle edit->redo command from menu</summary>
	public void OnEditRedo(object o, EventArgs args) 
	{
		if (dataBook.NPages == 0)
			return;
		DataView dv=((DataViewDisplay)dataBook.CurrentPageWidget).View;
		dv.Redo();
	}
	
	///<summary>Handle edit->cut command from menu</summary>
	public void OnEditCut(object o, EventArgs args) 
	{
		if (dataBook.NPages == 0)
			return;
		DataView dv=((DataViewDisplay)dataBook.CurrentPageWidget).View;
		dv.Cut();
	}
	
	///<summary>Handle edit->copy command from menu</summary>
	public void OnEditCopy(object o, EventArgs args) 
	{
		if (dataBook.NPages == 0)
			return;
		DataView dv=((DataViewDisplay)dataBook.CurrentPageWidget).View;
		dv.Copy();
	}
	
	///<summary>Handle edit->paste command from menu</summary>
	public void OnEditPaste(object o, EventArgs args) 
	{
		if (dataBook.NPages == 0)
			return;
		DataView dv=((DataViewDisplay)dataBook.CurrentPageWidget).View;
		dv.Paste();
	}
	
	///<summary>Handle edit->delete command from menu</summary>
	public void OnEditDelete(object o, EventArgs args) 
	{
		if (dataBook.NPages == 0)
			return;
		DataView dv=((DataViewDisplay)dataBook.CurrentPageWidget).View;
		dv.Delete();
	}
	
	///<summary>Handle the View->Layouts command</summary>
	public void OnViewLayouts(object o, EventArgs args)
	{
		if (dataBook.NPages == 0)
			return;
	
		DataView dv=((DataViewDisplay)dataBook.CurrentPageWidget).View;
		// show the selection dialog
		new LayoutSelectionDialog(dv);
	}
	
	///<summary>Handle the View->Statusbar->Show command</summary>
	public void OnViewStatusbarShow(object o, EventArgs args)
	{
		if (ViewStatusbarShowMenuItem.Active) {
			StatusPaned.Visible=true;
			ViewStatusbarOffsetMenuItem.Sensitive=true;
			ViewStatusbarSelectionMenuItem.Sensitive=true;
			ViewStatusbarOverwriteMenuItem.Sensitive=true;
		}
		else {
			StatusPaned.Visible=false;
			ViewStatusbarOffsetMenuItem.Sensitive=false;
			ViewStatusbarSelectionMenuItem.Sensitive=false;
			ViewStatusbarOverwriteMenuItem.Sensitive=false;
		}
	}
	
	///<summary>Handle the View->Statusbar->Offset command</summary>
	public void OnViewStatusbarOffset(object o, EventArgs args)
	{
		if (ViewStatusbarOffsetMenuItem.Active)
			OffsetStatusbar.Visible=true;
		else
			OffsetStatusbar.Visible=false;
			
		AssignResizeGrip();
	}
	
	///<summary>Handle the View->Statusbar->Selection command</summary>
	public void OnViewStatusbarSelection(object o, EventArgs args)
	{
		if (ViewStatusbarSelectionMenuItem.Active)
			SelectionStatusbar.Visible=true;
		else
			SelectionStatusbar.Visible=false;
		
		AssignResizeGrip();
	}
	
	///<summary>Handle the View->Statusbar->Selection command</summary>
	public void OnViewStatusbarOverwrite(object o, EventArgs args)
	{
		if (ViewStatusbarOverwriteMenuItem.Active)
			OverwriteStatusbar.Visible=true;
		else
			OverwriteStatusbar.Visible=false;
		
		AssignResizeGrip();
	}
	
	///<summary>Handle the Search->Find command</summary>
	public void OnSearchFind(object o, EventArgs args)
	{
		// get pattern from current selection...
		if (dataBook.NPages > 0) {
			DataView dv=((DataViewDisplay)dataBook.CurrentPageWidget).View;
			findDialog.LoadWithSelection(dv);
		}
		
		findDialog.Show();
	}
	
	///<summary>Handle the Search->Find Next command</summary>
	public void OnSearchFindNext(object o, EventArgs args)
	{
		if (!finder.FindNext()) {
			InformationAlert ia=new InformationAlert("The pattern you requested was not found.", "", MainWindow);
			ia.Run();
			ia.Destroy();
		}	
	}
	
	///<summary>Handle the Search->Find Previous command</summary>
	public void OnSearchFindPrevious(object o, EventArgs args)
	{
		if (!finder.FindPrevious()) {
			InformationAlert ia=new InformationAlert("The pattern you requested was not found.", "", MainWindow);
			ia.Run();
			ia.Destroy();
		}	
	}
	
	///<summary>Handle the Search->Find Previous command</summary>
	public void OnSearchReplace(object o, EventArgs args)
	{
		// get search pattern from current selection...
		if (dataBook.NPages > 0) {
			DataView dv=((DataViewDisplay)dataBook.CurrentPageWidget).View;
			replaceDialog.LoadWithSelection(dv);
		}
		
		replaceDialog.Show();
	}
	
	///<summary>Handle the Search->Find Next command</summary>
	public void OnSearchGotoOffset(object o, EventArgs args)
	{
		gotoOffsetDialog.Show();
	}
	
	///<summary>Handle the Help->About command</summary>
	public void OnHelpAbout(object o, EventArgs args)
	{
		AboutDialog.Show();
	}
	
	///<summary>Handle Main Window Delete Event</summary>
	public void OnMainWindowDeleteEvent (object o, DeleteEventArgs args) 
	{
		if (AskForSaveIfFilesChanged())
			Application.Quit ();
		args.RetVal = true;
	}

} // end bless main

