// created on 1/17/2005 at 5:00 PM
/*
 *   Copyright (c) 2005, Alexandros Frantzis (alf82 [at] freemail [dot] gr)
 *
 *   This file is part of Bless.
 *
 *   Bless is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   Bless is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Bless; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
 
using System;
using Gtk;
using Bless.Gui;
using Bless.Util;
using Bless.Gui.Dialogs;
using Bless.Buffers;
 
namespace Bless.Tools.Find {
                
public class FileFinder : IFinder
{
	DataBook dataBook;
	IFindStrategy strategy;
	Bless.Util.Range lastFound;
	
	public FileFinder(DataBook db)
	{ 
		// connect with databook
		dataBook=db;
		lastFound=new Bless.Util.Range();
	}

	public IFindStrategy Strategy 
	{ 
		get { return strategy;} 
		set { strategy=value; }
	}
	
	public Bless.Util.Range LastFound 
	{ 
		get { return lastFound;}
		set { lastFound=value; }
	}
	
	///<summary>Find the next occurence of the byte pattern</summary>
	public bool FindNext()
	{
		if (dataBook.NPages==0)
			return false;
			
		DataView dv=((DataViewDisplay)dataBook.CurrentPageWidget).View;
		
		strategy.Buffer=dv.Buffer;
		
		// decide where to start searching from
		if (!dv.Selection.IsEmpty())
			strategy.Position=dv.Selection.Start+1;
		else
			strategy.Position=dv.CursorOffset;
		
		Bless.Util.Range r=strategy.FindNext();
		
		// if pattern is found, select it and make it visible
		if (r!=null) {
			lastFound=r;
			dv.SetSelection(r.Start, r.End);//System.Console.WriteLine("Found at {0}-{1}", r.Start, r.End);
			dv.MoveCursor(r.End+1, 0);
			dv.Display.MakeOffsetVisible(r.Start, DataViewDisplay.ShowType.Closest);
			return true;
		}
		else {
			lastFound.Clear();
			return false;
		}
			
	}
	
	///<summary>Find the previous occurence of the byte pattern</summary>
	public bool FindPrevious()
	{
		if (dataBook.NPages==0)
			return false;
			
		DataView dv=((DataViewDisplay)dataBook.CurrentPageWidget).View;
		
		strategy.Buffer=dv.Buffer;
		
		// decide where to start searching from
		if (!dv.Selection.IsEmpty())
			strategy.Position=dv.Selection.End-1;
		else
			strategy.Position=dv.CursorOffset;
			
		
		Bless.Util.Range r=strategy.FindPrevious();
		
		// if pattern is found, select it and make it visible
		if (r!=null) {
			lastFound=r;
			dv.SetSelection(r.Start, r.End);//System.Console.WriteLine("Found at {0}-{1}", r.Start, r.End);
			dv.MoveCursor(r.End+1, 0);
			dv.Display.MakeOffsetVisible(r.Start, DataViewDisplay.ShowType.Closest);
			return true;
		}
		else {
			lastFound.Clear();
			return false;
		}
	
	}
	
	///<summary>Check if the specified range of bytes in a ByteBuffer equal a byte pattern</summary>
	bool RangeEqualsPattern(ByteBuffer bb, Bless.Util.Range sel, byte[] pattern)
	{
		int i=0;
		int len=pattern.Length;
		
		if (sel.IsEmpty())
			return false;
		
		while (i<len && pattern[i]==bb[sel.Start+i]) 
			i++;
		
		if (i!=len)
			return false;
		else
			return true;
		
	}
	
	///<summary>Replace the last match with a pattern</summary>
	public bool Replace(byte[] ba)
	{
		if (dataBook.NPages==0)
			return false;
			
		DataView dv=((DataViewDisplay)dataBook.CurrentPageWidget).View;
		if (RangeEqualsPattern(dv.Buffer, dv.Selection, strategy.Pattern)) {
			dv.Buffer.Replace(dv.Selection.Start, dv.Selection.End, ba);
			
			dv.CursorUndoStack.Push(new CursorState(dv.Selection.Start, 0, dv.Selection.Start+ba.Length, 0));
			dv.CursorRedoStack.Clear();
			
			dv.MoveCursor(dv.Selection.Start+ba.Length, 0);
			dv.SetSelection(-1, -1);
			
			lastFound.Clear();
			
			return true;
		}
		
		return false;
	}
	
	///<summary>Replace all matches with a pattern</summary>
	public int ReplaceAll(byte[] ba)
	{
		if (dataBook.NPages==0)
			return 0;
			
		DataView dv=((DataViewDisplay)dataBook.CurrentPageWidget).View;
		
		// initialize strategy
		strategy.Buffer=dv.Buffer;
		strategy.Position=0;
		
		// reset DataView
		dv.SetSelection(-1, -1);
		dv.MoveCursor(0, 0);
		
		int numReplaced=0;
		
		Bless.Util.Range match;
		Bless.Util.Range firstMatch=null;
		
		dv.Buffer.BeginActionChaining();
		
		// while we are finding matches, replace
		while ((match=strategy.FindNext())!=null) {
			if (firstMatch==null)
				firstMatch=match;	
			lastFound=match;
			
			dv.Buffer.Replace(match.Start, match.End, ba);
			
			numReplaced++;
		}
		
		dv.Buffer.EndActionChaining();
		
		dv.CursorUndoStack.Push(new CursorState(firstMatch.Start, 0, lastFound.Start+ba.Length, 0));
		
		// move cursor after final replacement
		dv.MoveCursor(lastFound.Start+ba.Length, 0);
		dv.Display.MakeOffsetVisible(lastFound.Start+ba.Length, DataViewDisplay.ShowType.Closest);
		
		return numReplaced;
	}
}
 
} // end namespace
