/*
 *   Copyright (c) 2004, Alexandros Frantzis (alf82 [at] freemail [dot] gr)
 *
 *   This file is part of Bless.
 *
 *   Bless is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   Bless is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Bless; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
using NUnit.Framework;
using System;

namespace Bless.Util {

///<summary>A double-linked list</summary>
public class List {

	///<summary>A node of the list</summary>
	public class Node {
		public object data;
		public Node next;
		public Node prev;
	
		public Node(object d, Node p, Node n) {
			data=d;
			prev=p;
			next=n;
		}
		
	}


	Node m_first;
	Node m_last;
	
	int m_count;
	
	
	public List() 
	{
		m_first=null;
		m_last=null;
	}
	
	
	private Node AddFirst(object o) 
	{
		m_first = new Node(o, null, null);
		m_last = m_first;
		++m_count;
		return m_first;
	}
	
	///<summary>Append an object to the list</summary>
	public Node Append(object o) 
	{
		return InsertAfter(m_last, o);
	}
	
	///<summary>Insert an object before a node in the list</summary>
	public Node InsertBefore(Node n, object o) 
	{
		if (m_last==null) { // first entry?
			return AddFirst(o);
		}	
		else {
			Node tmp = new Node(o, n.prev, n);
			if (n.prev!=null) 
				n.prev.next=tmp;
			n.prev=tmp;
			if (ReferenceEquals(n, m_first)) 
				m_first=tmp;
			++m_count;
			return tmp;
		}
	}
	
	///<summary>Insert an object after a node in the list</summary>
	public Node InsertAfter(Node n, object o) 
	{
		if (m_last==null) { // first entry?
			return AddFirst(o);
		}	
		else {
			Node tmp = new Node(o, n, n.next);
			if (n.next!=null)
				n.next.prev=tmp;
			n.next=tmp;
			if (ReferenceEquals(n, m_last)) 
				m_last=tmp;
			++m_count;
			return tmp;
		}
	}

	///<summary>Remove a node (unlink it)</summary>
	public void Remove(Node n) 
	{
		if (n.prev!=null) 
			n.prev.next=n.next;
		if (n.next!=null)
			n.next.prev=n.prev;
		if (ReferenceEquals(n, m_first)) 
			m_first=n.next;
		if (ReferenceEquals(n, m_last)) 
			m_last=n.prev;
		n.next=null;
		n.prev=null;
		--m_count;
	}
	
	public object this[int index] {
		get {
			//Console.WriteLine("Asking for index {0}", index);
			if (index > m_count)
				return null; 
			Node n=m_first;
			for(int i=0;i<index; i++)
				n=n.next;
			return n.data;
		  }
	}

	public void Display() 
	{
		Node n=m_first;
		while (n!=null) {
			Console.Write("{0}<=>",n.data);
			n=n.next;
		}
		Console.WriteLine();
	}
	
	public Node First {
		get { return m_first; }
	}

	public Node Last {
		get { return m_last; }
	}

	public int Count {
		get { return m_count; }
	}


}


#region Tests

[TestFixture]
public class ListTests {

	[Test]
	public void AppendTest() {
		List list = new List();
		list.Append(4);
		list.Append(7);
		List.Node n=list.First;
		Assert.AreEqual(4, (int)n.data);
		Assert.AreEqual(7, (int)n.next.data);
	}
	
	
	[Test]
	public void CountTest() {
		List list = new List();
		list.Append(1);
		list.Append(2);
		list.Append(3);

		Assert.AreEqual(3, list.Count);
	}

	
	[Test]
	public void InsertBeforeTest() {
		List list = new List();
		list.InsertBefore(list.First, 3);
		list.InsertBefore(list.First, 1);
		list.InsertBefore(list.Last, 2);

		
		List.Node n=list.First;
		Assert.AreEqual(1, (int)n.data);
		Assert.AreEqual(2, (int)n.next.data);
		Assert.AreEqual(3, (int)n.next.next.data);
	}

	[Test]
	public void InsertAfterTest() {
		List list = new List();
		list.InsertAfter(list.First, 1);
		list.InsertAfter(list.Last, 3);
		list.InsertAfter(list.First, 2);

		
		List.Node n=list.First;
		Assert.AreEqual(1, (int)n.data);
		Assert.AreEqual(2, (int)n.next.data);
		Assert.AreEqual(3, (int)n.next.next.data);
	}


	[Test]
	public void IndexerTest() {
		List list = new List();
		list.Append(1);
		list.Append(2);
		list.Append(3);

		Assert.AreEqual(1, list[0]);
		Assert.AreEqual(2, list[1]);
		Assert.AreEqual(3, list[2]);
		//if (list[3]!=null) 
			//Assert.Fail("Expected: null!");
	}

	[Test]
	public void RemoveTest() {
		List list = new List();
		List.Node n1=list.Append(1);
		list.Append(2);
		List.Node n3=list.Append(3);
		List.Node n4=list.Append(4);
		
		list.Remove(n1);
		Assert.AreEqual(3, list.Count,"#1");
		Assert.AreEqual(2, list[0],"#2");
		Assert.AreEqual(3, list[1],"#3");
		Assert.AreEqual(4, list[2],"#4");
		
		list.Remove(n3);
		Assert.AreEqual(2, list.Count,"#5");
		Assert.AreEqual(2, list[0],"#6");
		Assert.AreEqual(4, list[1],"#7");
		
		list.Remove(n4);
		Assert.AreEqual(1, list.Count,"#8");
		Assert.AreEqual(2, list[0],"#9");
		
		//if (list[3]!=null) 
			//Assert.Fail("Expected: null!");
	}

	[Test]
	public void RemoveLastTest() {
		List list = new List();
		List.Node n1=list.Append(1);
		List.Node n2=list.Append(2);
		
		list.Remove(n2);
		List.Node n3=list.Append(3);
		
		Assert.AreEqual(2, list.Count,"#1");
		Assert.AreEqual(1, list[0],"#2");
		Assert.AreEqual(3, list[1],"#3");
		
	}


}

#endregion
}
